/*
Copyright (c) 2013 General Development Systems

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "insd.hh"

#include <string>
#include <sys/types.h>
#include <sys/socket.h>

namespace insd {

bool Connection::active() {
    return !masterInactive() && connected;
}

std::string Connection::get(int nchar) {
    if(!connected)
        throw ConnectionTerminatedException();

    if(masterInactive()) { //shall return false if there is no master
        connected = false;
        throw ConnectionTerminatedException();
    }

    timeval timeout;
    fd_set selectedFDs; //select() requires a list of socket FDs in this format.
    while(connected && !masterInactive()) {

        //Fill the selectedFDs set with our server socket and new client sockets
        //(a call to select() changes the set, so we need to do this every time)
        FD_ZERO(&selectedFDs);
        FD_SET(clientFD, &selectedFDs); //add the server FD to the selectedFDs list
        int maxFD=clientFD;

        //specify a timeout so an absense of new connection activity doesn't make the loop hang.
        //calling select() changes these values, so need to redefine each iteration.
        timeout.tv_sec = 0; timeout.tv_usec = 50000; //50ms

        //call select() which will return either when a socket is ready to read, or
        //after a timeout.
        int numReady = select(maxFD+1, &selectedFDs, NULL, NULL, &timeout);
        if(numReady == -1) { //error (e.g. interrupt)
            break;
        }

        if(numReady > 0 && FD_ISSET(clientFD, &selectedFDs)) { //check if the event returned was from clientFD
            //allocate a character array large enough to buffer the incoming characters...
            char buff[nchar+1]; //allocate it like this so it's freed if an exception's thrown.

            int nrecv = recv(clientFD, buff, nchar, 0);

            if(nrecv <= 0) {
                //something wrong!
                connected = false;
                throw ConnectionTerminatedException();
            } else {
                buff[nrecv] = '\0'; //null terminate the received string
            }
            return std::string((const char*)buff);
        }
    }
    connected = false;
    throw ConnectionTerminatedException();
}

void Connection::put(std::string str) {
    if(!connected)
        throw ConnectionTerminatedException();

    if(send(clientFD, str.c_str(), str.size(), 0) == -1) {
        connected = false;
        throw ConnectionTerminatedException();
    }
}

Connection::Connection(int fd, SocketServ* creator) {
    clientFD = fd;
    this->master = creator;
    connected = true;
}

Connection::~Connection() {
    //close the connection
    if(close(clientFD) == -1) { //This doesn't seem to always work UNLESS we check the return value.
        throw std::system_error(errno, std::generic_category(), "insd closing connection");
    }
}

bool Connection::masterInactive() {
//returns true in the server process if the master SocketServ object has stopped.
//This is a signal that all connections on the server side should be terminated and their
//threads rejoined.
//In a client process, this always returns false.
    if(master) {
        return !master->active();
    }
    return false;
}

}

