/*
Copyright (c) 2013 General Development Systems

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "insd.hh"

#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <sys/un.h>
#include <fcntl.h>
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <string>

namespace insd {
    
PidFileHolder* pfh;

ServerInstance::ServerInstance(std::string workingDir, std::string processName) {
    pfh = nullptr;
    this->workingDir = workingDir;
    this->pidFile = processName + ".pid";
    this->logFile = processName + ".log";

    //evaluate what pidFileFullPath should be...
    if(pidFile[0] == '/') { //absolute path
        this->pidFileFullPath = pidFile;
    } else { //relative path.
        this->pidFileFullPath = workingDir + pidFile;
    }
}

void ServerInstance::run(ServerMain& serverMain, bool makeLog) {
//void type because the forked server process doesn't return -- it exits with an appropriate error code however.
//We need to encapsulate the bulk of this function into a separate function, runAux, because
//otherwise calling an exit() directly will fail to call cleanup destructors.
//Instead, an exception insd::Exit is provided which pops the stack up to this level and calls
//destructors of automatic variables along the way.
    try {
        runAux(serverMain, makeLog);
    } catch(Exit e) { //only caught in child process.
        exit(e);
    }
}

void exitHandler() {
//bound to an invocation of the exit() funtion within the child process.
    if(pfh) {
        delete pfh;
        pfh = nullptr;
    }
}


void ServerInstance::runAux(ServerMain& serverMain, bool makeLog) {
    if(chdir(workingDir.c_str())==-1) //set working directory
        throw(DirectoryError(errno)); //could not set directory

    umask(0); //Linux daemon writing howto recommends setting this for file access.

    //check if a server instance is already running...
    if(isRunning()) throw(LockFileExistsException());

    pid_t pid = fork(); //***********************************FORK HAPPENS HERE
    if(pid < 0) { //unable to fork new process.
        throw std::system_error(errno, std::generic_category(), "insd fork syscall");
    } else if(pid > 0) { //inside parent process.
        return;
    }
    //***************Everything from here happens inside new child (server) process.

    atexit(exitHandler);

    pfh = new PidFileHolder(pidFile);
    //Create the pidfile, which is always required. It will automatically be deleted
    //when the PidFileHolder object is destroyed (should be deleted as an atexit task)

    int errnum;
    errnum = serverMain.init();
    if(errnum) {
        throw Exit(errnum); //nonzero errnum indicates init failure
    }

    for(int i = 0; i<=1; i++) close(i); //close stdio descriptors for safety. 
    {
    //open and redirect standard i/o descriptors to null for safety.
        int i = open("/dev/null", O_RDWR); //stdin
        if(i==-1) throw std::system_error(errno, std::generic_category(), "insd ServerInstance::run()");
        if(dup(i)==-1) throw std::system_error(errno, std::generic_category(), "insd ServerInstance::run()"); //stdout
        if(!makeLog) { //stderr can either go to /dev/null, or to a logfile.
            close(2);
            if(dup(i)==-1) throw std::system_error(errno, std::generic_category(), "insd ServerInstance::run()"); //stderr
        } else {
            freopen(logFile.c_str(), "a", stderr);
        }
    }
    if(setsid()==-1) //detach from terminal
            throw std::system_error(errno, std::generic_category(), "insd ServerInstance::run() setsid call");

    errnum = serverMain(); //run the server, returns here when done.
    throw Exit(errnum); //the forked process has no appropriate calling function to return to.
}

pid_t ServerInstance::isRunning() { //todo: check process table to make sure the pid is correct.
                            //todo: ensure the file actually contains an integer.
    return PidFileHolder::conflictExists(pidFileFullPath);
}

bool ServerInstance::stop(int timeout) {
//stop an already-running daemon process by checking the pidfile.
//If timeout (in seconds) is nonzero, the function will wait until
//timeout for the pidfile to disappear and return true on success
//or false if the daemon hasn't responded.
    pid_t thePid = isRunning();
    if(!thePid) throw(NotRunningException());
    kill(thePid, SIGTERM);

    for(int t=0; t<timeout; t++) {
        sleep(1);
        if(!isRunning()) return true;
    }
    return false;
}

void ServerInstance::destroy() {
    pid_t thePid = isRunning();
    if(!thePid) throw(NotRunningException());
    kill(thePid, SIGKILL);

    unlink(pidFileFullPath.c_str());
}

}//namespace

