/*
Copyright (c) 2013 General Development Systems

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "insd.hh"

#include <iostream>
#include <string>
#include <cstring>
#include <condition_variable>
#include <sys/socket.h>
#include <sys/un.h>
#include <signal.h>
#include <unistd.h>

namespace insd {

SocketServ* SocketServ::signalee = nullptr;

SocketServ::SocketServ() {
    signal(SIGPIPE, SIG_IGN); //ignore SIGPIPEs which may otherwise crash the client/server silently.

    //create new server socket to listen for new connections...
    keepRunning = false;
    filemapped = false;
    threadCount = 0;
    serverFD = socket(AF_UNIX, SOCK_STREAM, 0);
    if(serverFD == -1) {
        throw std::system_error(errno, std::generic_category(), "insd"); 
    }
}

SocketServ::~SocketServ() {
    stop();
    close(serverFD);
    if(filemapped) {
        unlink(sockfile.c_str()); //delete any existing socket file
    }

    //Wait for all threads to terminate
    std::unique_lock<std::mutex> ul(threadCountMutex);
    threadCountCondVar.wait(ul, [this]{return threadCount <= 0;});
}

void SocketServ::mapFile(std::string filename) {
//create socket file bound to serverFD...
    struct sockaddr_un sockServAddr;
    sockfile = filename;
    sockServAddr.sun_family = AF_UNIX;
    strcpy(sockServAddr.sun_path, filename.c_str());
    unlink(sockServAddr.sun_path); //delete any existing socket file
    if(bind(serverFD, (struct sockaddr*)&sockServAddr, sizeof(struct sockaddr_un)) == -1) {
        throw(CouldNotMapFileException()); //could not bind socket to file
    }
    filemapped = true;

    //there has to be a better place to put this!!
    if(listen(serverFD, 15 /*max connections in accept queue*/) == -1) {
        throw std::system_error(errno, std::generic_category(), "insd"); 
    }
}


void SocketServ::operator()(NewConHandler& cb) {
//This is an 'infinite' loop that keeps monitoring for new connections
//until a program interrupt is received, or stop() is called from a different thread.

    //redirect signals, so that (for example) a ^C will terminate the SocketServ main loop function
    //and return control to the caller (to do cleanup) rather than terminating abruptly.
    //In future, change this to make it more graceful: to redirect signals back again after the function terminates.

    signal(SIGTERM, sigHandler);
    signal(SIGINT, sigHandler);

    keepRunning = true;

    timeval timeout;
    fd_set selectedFDs; //select() requires a list of socket FDs in this format.
    while(active()) {
        //Fill the selectedFDs set with our server socket and new client sockets
        //(a call to select() changes the set, so we need to do this every time)
        FD_ZERO(&selectedFDs);
        FD_SET(serverFD, &selectedFDs); //add the server FD to the selectedFDs list
        int maxFD=serverFD;

        //specify a timeout so an absense of new connection activity doesn't make the loop hang.
        //calling select() changes these values, so need to redefine each iteration.
        timeout.tv_sec = 1; timeout.tv_usec = 0; 

        //call select() which will return either when a socket is ready to read, or
        //after a timeout.
        int numReady = select(maxFD+1, &selectedFDs, NULL, NULL, &timeout);
        if(numReady == -1) { //error (e.g. interrupt)
            break;
        }

        //now to deal with any event(s) select() has returned...

        //first check if the serverFD is ready with a new connection to accept...
        while(numReady > 0 && FD_ISSET(serverFD, &selectedFDs)) {
            struct sockaddr_un sockClientAddr;
            socklen_t t = sizeof(sockClientAddr);
            int clientDescriptor = accept(serverFD, (struct sockaddr *)&sockClientAddr, &t);
            if(clientDescriptor == -1) {
		throw std::system_error(errno, std::generic_category(), "insd accepting client connection"); 
            }
            //we now have a new open connection. Notify callback.
            //auto thd = new std::thread(runNewConThread, &cb, clientDescriptor, this);
            try {
                std::thread thd(runNewConThread, &cb, clientDescriptor, this);
                thd.detach();
            } catch(std::system_error e) {
                shutdown(clientDescriptor, SHUT_RDWR); 
                //terminate the new connection we can't handle right now.
            }
            //clientThreads.push_back(thd);

            FD_CLR(serverFD, &selectedFDs); //remove event from the set
            numReady--; //one less event to worry about.
        }
    }
}


void SocketServ::runNewConThread(NewConHandler* handlerObject, int clientFD, SocketServ* instance) {
    std::unique_lock<std::mutex> ul(instance->threadCountMutex);
    instance->threadCount++;
    ul.unlock();

    { //encapsulated scope to hopefully call destructor early!
        Connection c(clientFD, instance);
        (*handlerObject)(c); 
    }

    //at this point, we should notify of thread termination.
    ul.lock();
    instance->threadCount--;
    //std::notify_all_at_thread_exit(instance->threadCountCondVar, ul); //NOT SUPPORTED BY GCC!!
    instance->threadCountCondVar.notify_all();                       //next best thing!
}

void SocketServ::stop() {
    std::lock_guard<std::mutex> lg(keepRunningGuard);
    keepRunning = false;
}

void SocketServ::sigHandler(int sig) {
//static handler function for SIGINT and SIGTERM
    std::cerr << "insd: received signal " << sig << std::endl;
    if(signalee)
        signalee->stop();
}


bool SocketServ::active() {
    std::lock_guard<std::mutex> lg(keepRunningGuard);
    return keepRunning;
}


} //namespace

