/*
Copyright (c) 2013 General Development Systems

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <mutex>
#include <condition_variable>
#include <list>
#include <thread>
#include <exception>
#include <system_error>


namespace insd {

class SocketServ;
class Connection;

class LockFileExistsException {};       //thrown when creating a new daemon process if the lock file already exists
class NotRunningException{};    //thrown when server process is not running
class DirectoryError : public std::system_error { //thrown when we are unable to change to the requested working directory
public:
    DirectoryError(int errnum) : std::system_error(errnum, std::generic_category(), "insd") {}
};
class ConnectionTerminatedException {}; //thrown when a client connection has been closed
class CouldNotMapFileException {};      //thrown when creating socket server
class Exit {                           //throw to exit the server.
public:
    int val;
    Exit(int errcode) {val = errcode; }
    operator int() {return val;}
};

Connection* clientConnect(std::string); //in socketClient.cc

class ServerMain {
//user must implement an object from this abstract class to run the server daemon.
public:
    virtual int init() {return 0;} //override for initialisation to be done before detaching from terminal
                                   //must return zero on success.
    virtual int operator()()=0;
};

class PidFileHolder {
public:
    class ExistingConflictException : public std::exception {
    public:
        ExistingConflictException(pid_t conflictingProcess) {;}
    };

    PidFileHolder(std::string filepath);
    ~PidFileHolder();

    static pid_t conflictExists(std::string filepath);
private:
    std::string filepath;
};

class ServerInstance {
public:
    ServerInstance(std::string runPath, std::string processName);
    void run(ServerMain& handler, bool makeLog);
    bool stop(int timeout); //send TERM signal to the pid stored in the pidfile.
    void destroy(); //send KILL signal to the pid stored in the pidfile, and delete the pidfile.
    pid_t isRunning(); //returns the pid of an already-running server process, or 0 if not running.
private:
    void runAux(ServerMain& handler, bool makeLog);
    std::string workingDir;
    std::string pidFile;
    std::string logFile;
    std::string pidFileFullPath; //rename to pidFileFullPath
};

class Connection {
//used by server and client-side connection instances.
public:
    ~Connection();

    bool active(); //returns true iff the connection is active.

    std::string get(int size = 1);
    //receive (with blocking wait) up to the specified number of characters. May throw a ConnectionTerminatedException 
    
    void put(std::string);
    //sends the contents of the string across the connection in one go
protected:
    Connection(int fdescriptor, SocketServ* creator); //protected constructor used for serverside creation
    int clientFD;
    SocketServ* master;

    friend class SocketServ;
    friend Connection* clientConnect(std::string);
private:
    bool connected;
    bool masterInactive();
};




class NewConHandler {
//user must implement this abstract class and pass such an object to the SocketServ function object.
//the instance will be called each time a new connection is available.
public:
    virtual void operator()(Connection&) =0;
};


class SocketServ {
public:

    SocketServ();
    ~SocketServ();

    void mapFile(std::string filename); //attach a file to the socket.
    //An existing file of the same name will be overwritten, and the file will
    //be deleted when the SocketServ object is destroyed.

    bool active();
    //returns true if the socket should still be active. If it becomes false, 
    //it provides an instruction for the server and all connections to cease
    //honoring their instructions (begin returning errors instead), terminate 
    //their connections and clean up.

    void operator()(NewConHandler&);
    void stop();
    static void sigHandler(int);
private:
    static SocketServ* signalee; //a pointer to the instance of SocketServ that sigHandler should act on when it is invoked to handle a program SIGTERM or SIGINT.

    int serverFD;

    bool filemapped;
    bool keepRunning;
    std::mutex keepRunningGuard;

    std::string sockfile;

    static void runNewConThread(NewConHandler*, int clientFD, SocketServ* instance);
    std::list<std::thread*> clientThreads;

    int threadCount; //Keep track of active threads/connections, so we ensure threads have all terminated before we do.
    std::mutex threadCountMutex;
    std::condition_variable threadCountCondVar; //threads provide notification of having decremented thread count.

};


}
